/*

    sftpd - meant to be a small and flexible ftp server implementation
    Copyright (C) 2009  Jan Murawski  der@oelgoetze.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

    For more information about the GNU Public license see:
    http://www.gnu.org/copyleft/gpl.html

*/

#include "serverObj.h"


//
// init server object whith default values
//
ServerObj::ServerObj()
{
    InitializeCriticalSection(&csl);
    controlePort = 21;
    dataPort = 20;
    listenning = false;
    getFinished = false;
}

ServerObj::~ServerObj()
{
    DeleteCriticalSection(&csl);
}


//
// add a new user account to our list
//
void ServerObj::addUser(char *user, char *password, char *folder, bool folderWritable)
{
    EnterCriticalSection(&csl);

    ftpUser *newUser = new ftpUser;
    int pos;
    
    //add all chars to string
    pos = 0;
    while (*(user + pos) != 0)
        newUser->user += *(user + pos++);

    pos = 0;
    while (*(password + pos) != 0)
        newUser->password += *(password + pos++);

    pos = 0;
    while (*(folder + pos) != 0)
        newUser->folder += *(folder + pos++);

    newUser->folderWritable = folderWritable;

    // add user to list
    users.push_back(newUser);

    LeaveCriticalSection(&csl);
}


//
// test, if given user name is in our list
//
bool ServerObj::userExists(string user)
{
    EnterCriticalSection(&csl);

    for (unsigned int i = 0; i < users.size(); i++)
        if ((*users.at(i)).user == user)
        {
            LeaveCriticalSection(&csl);
            return true;
        }

    LeaveCriticalSection(&csl);
    return false;
}

//
// test, if given username and password do match an entry in our list
//
bool ServerObj::loginValid(string user, string password)
{
    EnterCriticalSection(&csl);

    for (unsigned int i = 0; i < users.size(); i++)
        if ((*users.at(i)).user == user)
        {
            if ((*users.at(i)).password == password)
            {
                LeaveCriticalSection(&csl);
                return true;
            }
            else
                break;
        }

    LeaveCriticalSection(&csl);
    return false;
}


//
// return dircetory and rw-settings, if login is valid
//
bool ServerObj::getLoginSettings(string user, string password, string &directory, string &rw)
{
    EnterCriticalSection(&csl);

    for (unsigned int i = 0; i < users.size(); i++)
        if ((*users.at(i)).user == user)
        {
            if ((*users.at(i)).password == password)
            {
                directory = (*users.at(i)).folder;
                rw = (*users.at(i)).folderWritable;
                LeaveCriticalSection(&csl);
                return true;
            }
            else
                break;
        }

    LeaveCriticalSection(&csl);
    return false;
}


//
// change the port values (needs restart)
//
bool ServerObj::setPorts(int controlePort, int dataPort)
{
    EnterCriticalSection(&csl);

    // cannot change ports while running
    if (listenning)
    {
        LeaveCriticalSection(&csl);
        return false;
    }
    
    // ports in valid range?
    if (!(controlePort > 0 && controlePort <= 65535 
        && dataPort > 0 && dataPort <= 65535))
    {
        LeaveCriticalSection(&csl);
        return false;
    }
    
    LeaveCriticalSection(&csl);
    return true;
}


//
// start the server
//
bool ServerObj::start()
{
    EnterCriticalSection(&csl);

    // server already running?
    if (listenning)
    {
        LeaveCriticalSection(&csl);
        return true;
    }

    // init winsock
    WSADATA wsa;
    if (WSAStartup(MAKEWORD(2,0), &wsa) != 0)
    {
        WSACleanup();
        // FIXME: INSERT ERROR LOGGING (!)
        LeaveCriticalSection(&csl);
        return false;
    }
    
    // create controle port listen socket and bind it to defined port
    // no listenning here, just binding
    SOCKET *ctrlListenSock = new SOCKET;
    {
        *ctrlListenSock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
        if(*ctrlListenSock == INVALID_SOCKET)
        {
            WSACleanup();
            // FIXME: INSERT ERROR LOGGING (!)
            LeaveCriticalSection(&csl);
            return false;       
        }

        SOCKADDR_IN serverInf;
        serverInf.sin_family        = AF_INET;
        serverInf.sin_addr.s_addr   = INADDR_ANY;
        serverInf.sin_port          = htons(controlePort);
        if(bind(*ctrlListenSock,(SOCKADDR *)(&serverInf), sizeof(serverInf)) == SOCKET_ERROR)
        {
            WSACleanup();
            // FIXME: INSERT ERROR LOGGING (!)
            LeaveCriticalSection(&csl);
            return false;   
        }
    }
    
    // create data port listen socket and bind it to defined port
    // no listenning here, just binding
    SOCKET *dataListenSock = new SOCKET;
    {
        *dataListenSock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
        if(*dataListenSock == INVALID_SOCKET)
        {
            WSACleanup();
            // FIXME: INSERT ERROR LOGGING (!)
            LeaveCriticalSection(&csl);
            return false;       
        }

        SOCKADDR_IN serverInf;
        serverInf.sin_family        = AF_INET;
        serverInf.sin_addr.s_addr   = INADDR_ANY;
        serverInf.sin_port          = htons(dataPort);
        if(bind(*dataListenSock,(SOCKADDR *)(&serverInf), sizeof(serverInf)) == SOCKET_ERROR)
        {
            WSACleanup();
            // FIXME: INSERT ERROR LOGGING (!)
            LeaveCriticalSection(&csl);
            return false;   
        }
    }


    // init went fine so let's start the thread to handle this connection
    listenThreadInfo *lti = new listenThreadInfo;
    lti->svObj = this;
    lti->ctrlListenSock = ctrlListenSock;
    lti->dataListenSock = dataListenSock;

    {
        DWORD threadId;
        HANDLE h = CreateThread(NULL, 0,  &listenThread, lti, 0, &threadId);
        regThread(h);
    }

    LeaveCriticalSection(&csl);
    return true;
}


//
// should the server be stopped?
//
bool ServerObj::getGetFinished()
{
    EnterCriticalSection(&csl);

    bool retVal = getFinished;

    LeaveCriticalSection(&csl);
    return retVal;
}


//
// add a new thread to our list, so that we know what is running
//
void ServerObj::regThread(HANDLE hThread)
{
    EnterCriticalSection(&csl);

    threadList.push_back(hThread);

    LeaveCriticalSection(&csl);
    maintainThreadList();
}



//
// remove finished threads from list
//
void ServerObj::maintainThreadList()
{
    EnterCriticalSection(&csl);

    DWORD exitCode;
    HANDLE handle;
    for (unsigned int i = 0; i < threadList.size(); i++)
    {
        handle = threadList.at(i);
        if (!GetExitCodeThread(handle, &exitCode))
            continue;

        // remove entry if thread is done
        if (exitCode != STILL_ACTIVE)
        {
            CloseHandle(handle);
            threadList.erase(threadList.begin() + i);
            i--;
        }
    }

    LeaveCriticalSection(&csl);
}


//
// wait for all threads to stop
//
void ServerObj::stop()
{
    int items;

    getFinished = true;

    // wait for all associated threads to exit
    for (;;)
    {
        maintainThreadList();

        EnterCriticalSection(&csl);
        items = threadList.size();
        LeaveCriticalSection(&csl);

        if (items == 0)
            break;

        Sleep(1);
    }

    // get ready again
    EnterCriticalSection(&csl);
    listenning = false;
    getFinished = false;
    LeaveCriticalSection(&csl);
}


//
// returns date port
//
int ServerObj::getDataPort()
{
    return this->dataPort;
}


//
// register a new data connection request and return handle
//
dataConnHandle* ServerObj::createDataConnHandle(string ip)
{
    dataConnHandle *dch = new dataConnHandle;

    dch->remoteIp = ip;
    dch->sck = 0;
    dch->initializedAt = GetTickCount();
    dch->timedOut = false;

    EnterCriticalSection(&csl);
    dchList.push_back(dch);
    LeaveCriticalSection(&csl);

    return dch;
}


//
// delete all memory allocated for handle
//
void ServerObj::cleanUpDataConnHandle(dataConnHandle *dch)
{
    if (dch->sck != 0)
    {
        closesocket(*(dch->sck));
        delete dch->sck;
    }

    delete dch;
}


//
// returns first (!) handle, if request for given ip is existant
//
dataConnHandle* ServerObj::findDataConnHandle(string ip)
{
    dataConnHandle *retVal = 0;
    EnterCriticalSection(&csl);

    for (unsigned int i = 0; i < dchList.size(); i++)
        if (dchList.at(i)->remoteIp == ip)
        {
            retVal = dchList.at(i);
            break;
        }

    LeaveCriticalSection(&csl);
    return retVal;
}


//
// assign socket to handel. returns false if dch is no more existant or
// socket already assigned
//
bool ServerObj::registerSocketWithDCH(dataConnHandle *dch, SOCKET *sck)
{
    bool retVal = false;
    EnterCriticalSection(&csl);
    
    for (int i = dchList.size(); i > 0; i--)
        if (dchList.at(i -1) == dch)
        {
            if (dch-> sck == 0)
            {
                dch->sck = sck;
                retVal = true;
            }
            break;
        }

    LeaveCriticalSection(&csl);
    return retVal;
}


bool ServerObj::isSocketRegistered(dataConnHandle *dch)
{
    // assume, dch has not been removed - should not happen ^^
    
    bool retVal = false;
    EnterCriticalSection(&csl);

    if (dch->sck != 0)
        retVal = true;

    LeaveCriticalSection(&csl);
    return retVal;
}


//
// delete last data connection request entry matching given ip
// returns entry
//
dataConnHandle* ServerObj::deleteDataConnHandleEntry(string ip)
{
    dataConnHandle *retVal = 0;
    EnterCriticalSection(&csl);

    for (int i = dchList.size(); i > 0; i--)
        if (dchList.at(i -1)->remoteIp == ip)
        {
            retVal = dchList.at(i -1);
            dchList.erase(dchList.begin() + i -1);  // FIXME: index conversion correct !?
            break;
        }
    
    LeaveCriticalSection(&csl);
    return retVal;
}


//
// delete last data connection request entry matching given handle
// returns entry
//
dataConnHandle* ServerObj::deleteDataConnHandleEntry(dataConnHandle *dch)
{
    dataConnHandle *retVal = 0;
    EnterCriticalSection(&csl);

    for (int i = dchList.size(); i > 0; i--)
        if (dchList.at(i -1) == dch)
        {
            retVal = dchList.at(i -1);
            dchList.erase(dchList.begin() + i -1);  // FIXME: index conversion correct !?
            break;
        }
    
    LeaveCriticalSection(&csl);
    return retVal;
}


//
// reset timer for given date connection
//
bool ServerObj::resetDataConnHandleTimer(dataConnHandle *dch)
{
    bool retVal = false;
    EnterCriticalSection(&csl);

    for (int i = dchList.size(); i > 0; i--)
        if (dchList.at(i -1) == dch)
        {
            retVal = true;
            dchList.at(i -1)->initializedAt = GetTickCount();
            break;
        }
    
    LeaveCriticalSection(&csl);
    return retVal;
}


//
// check if any incomming connection is pending for to long now...
//
void ServerObj::maintainDataConnHandleList()
{
    EnterCriticalSection(&csl);

    for (unsigned int i = dchList.size(); i > 0; i--)
        if (dchList.at(i -1)->initializedAt + WAIT_FOR_NEW_DATA_CONN_TIMEOUT < GetTickCount())
        {
            dchList.at(i -1)->timedOut = true;
            break;
        }
    
    LeaveCriticalSection(&csl);
}


//
// return, if connection request hast timed out
//
bool ServerObj::isDataConnHandleTimedOut(dataConnHandle *dch)
{
    bool retVal = true;     // return true, if dch is no more in list
    EnterCriticalSection(&csl);

    for (int i = dchList.size(); i > 0; i--)
        if (dchList.at(i -1) == dch)
        {
            retVal = dchList.at(i -1)->timedOut;
            break;
        }
    
    LeaveCriticalSection(&csl);
    return retVal;
}


//
// returns number of all (waiting, timed out, registered) items in list
//
unsigned int ServerObj::getDataConnHandleCount()
{
    EnterCriticalSection(&csl);
    
    unsigned int retVal = dchList.size();
    
    LeaveCriticalSection(&csl);

    return retVal;
}
